/* Generic C includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

/* Structures, protos and erreurs of the lib */
#include "libregionMap.h"

#ifndef MAX
#define MAX(a,b) ((a) > (b) ? (a) : (b))
#endif
#ifndef MIN
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif

/**********************************************************************
 *
 * Fuses a 3D image in the regionMap
 *
 **********************************************************************/

int
rMap_fuseIm3d (REGION_MAP*	regionMap,
	       DATA_IM3D*	im3d,
	       T3D*		main2origin)

{
  int		i,j;		/* To loop on the 3D points */
  int		iRegion, jRegion; /* Coordinate of a region of the region map */
  DATA_PT3D*	pt3d;		/* A point of the im3d */
  double	xOrigin;
  double	yOrigin;	/* Coordinates of the top-left corner of the
				   first region of the region map */
  double	scale;		/* region map scale */
  RMAP_REGION*	region;
  int		currentDate;	/* The "date" corresponding to the current 3D
				   image */ 
  double	x, y;
  double	xReg, yReg;
  double	normalNorm;		/* To normalize the normal vector */

  /* Checks the parameters */
  if (im3d->header.coordonnees != COORDONNEES_CARTESIENNES)
    return rMapError (ERR_RMAP_IM3D);
  if (im3d->header.points != POINTS_CALCULES)
    return rMapError (ERR_RMAP_IM3D);

  t3dConvertTo(T3D_EULER, &(regionMap->origin2corner));
  xOrigin = regionMap->origin2corner.euler.euler[3];
  yOrigin = regionMap->origin2corner.euler.euler[4];
  scale = regionMap->scale;

  currentDate = regionMap->nbUpdates + 1;

  /* Allocate the local informations of the 3D image */
  if (alloue_im3d_infos_locales (im3d, sizeof(RMAP_REGION_INDICES)) != 0)
    return rMapError (ERR_RMAP_IM3D);

  /* Sets all the regions updateState to "RMAP_REGION_PERCEIVED", and resets the
     previously computed attributes */
  for (i = regionMap->last_iMin; i <= regionMap->last_iMax; i++) {
    for (j = regionMap->last_jMin; j <= regionMap->last_jMax; j++) {
      region = regionMap->regions + (i * regionMap->nbcol) + j;
      if ((region->updateState == RMAP_REGION_NEWLY_PERCEIVED)
	  || (region->updateState == RMAP_REGION_NOT_ENOUGH_POINTS))
	region->updateState = RMAP_REGION_PERCEIVED;
      if (region->lastDate == currentDate - 1) { /* Shouldn't we remove this
						    test ? TODO */
	region->infos3d.attributesState = NOT_COMPUTED;
	region->infos3d.nbNormales = 0;
	memset ((double*) region->infos3d.attributes, 0, 
		NB_3D_ATTRIBUTES * sizeof (double));
      }
    }
  }

  /* Resets the bounds of the last perceived area... */
  regionMap->last_iMin = regionMap->nblig;
  regionMap->last_iMax = 0;
  regionMap->last_jMin = regionMap->nbcol;
  regionMap->last_jMax = 0;
  /* ... and the number of newly perceived cells */
  regionMap->nbNewCells = 0;

  /* Now loops on the points of the 3D image. For each of the points: finds the
     region in which it is vertically projected and update the associated
     classification structure. */
  pt3d = im3d->points_tab;
  for (i = 0; i < im3d->header.nblig; i++) {
    for (j = 0; j < im3d->header.nbcol; j++) {
      if (pt3d->state == GOOD_PT3D) {
	iRegion = (pt3d->coord_1 - xOrigin) / scale;
	jRegion = (pt3d->coord_2 - yOrigin) / scale;
	  
	if ((iRegion < 0) || (iRegion >= regionMap->nblig)
	      || (jRegion < 0) || (jRegion >= regionMap->nbcol))
	  ;			/* The region is out of the region map */
	else {
	  region = regionMap->regions + (iRegion * regionMap->nbcol) + jRegion;

	  /* If this is the first time we have a point un this region for the
	     considered 3D image, update some global region and regionMap
	     characteristics  */
	  if (region->lastDate != currentDate)
	    rMap_updateStates (regionMap, region, iRegion, jRegion, currentDate);
	  else
	    /* The region has already been perceived -- update its lastDate */
	    region->lastDate = currentDate;

	  /* Now update the region 3D attributes */
	  rmap_update3dAttributes (&(region->infos3d), pt3d);

	  /* Update the pt3d local structure with the region indices */
	  pt3d->flag_perso = 1;
	  ((RMAP_REGION_INDICES*) pt3d->local_tab)->iRegion = iRegion;
	  ((RMAP_REGION_INDICES*) pt3d->local_tab)->jRegion = jRegion;
	}

      }
      pt3d++;
    }
  }

  /* Memorize the current robot pose in the position list of the regionMap */
  t3dConvertTo (T3D_EULER, main2origin);
  memcpy (&(regionMap->positions[regionMap->nbUpdates]), main2origin, sizeof (T3D));

  /* Loop on all the updated regions: 
     - to compute their distance wrt. the sensor position (this is done in a
     separate loop to avoid re-computing it every time a point is added in a
     region) 
     - to normalize the mean normal vector 
     - to finish the incremental computation of the variances
     - to check whether we have enough points to classify the region
  */ 
  x = main2origin->euler.euler[3];
  y = main2origin->euler.euler[4];
  for (i = regionMap->last_iMin; i <= regionMap->last_iMax; i++) {
    for (j = regionMap->last_jMin; j <= regionMap->last_jMax; j++) {
      region = regionMap->regions + (i * regionMap->nbcol) + j;
      if (region->lastDate == currentDate) {
	/* Compute the MEAN_H_DISTANCE attribute */
	xReg = (i + 0.5) * regionMap->scale 
	  + regionMap->origin2corner.euler.euler[3];
	yReg = (j + 0.5) * regionMap->scale 
	  + regionMap->origin2corner.euler.euler[4];
	region->infos3d.attributes[MEAN_H_DISTANCE] =
	  sqrt ( (xReg - x)*(xReg - x) + (yReg - y)*(yReg - y) ); 
	/* Normalize the normal vector */
	normalNorm = sqrt ((region->infos3d.attributes[NX]
			    * region->infos3d.attributes[NX])
			   + (region->infos3d.attributes[NY] 
			      * region->infos3d.attributes[NY])
			   + (region->infos3d.attributes[NZ] 
			      * region->infos3d.attributes[NZ]));
	region->infos3d.attributes[NX] =
	  region->infos3d.attributes[NX] / normalNorm; 
	region->infos3d.attributes[NY] =
	  region->infos3d.attributes[NY] / normalNorm; 
	region->infos3d.attributes[NZ] =
	  region->infos3d.attributes[NZ] / normalNorm; 
	/* Check the number of available points */
	if (region->infos3d.attributes[NB_PTS] < regionMap->nbPointsMin)
	  region->updateState = RMAP_REGION_NOT_ENOUGH_POINTS;
	else { /* Finish the variance computation */
	  region->infos3d.attributes[VAR_Z] 
	    /= (region->infos3d.attributes[NB_PTS] -1);
	  if (region->infos3d.nbNormales > 1) {
	    region->infos3d.attributes[VAR_NX] 
	      /= (region->infos3d.nbNormales -1);
	    region->infos3d.attributes[VAR_NY] 
	      /= (region->infos3d.nbNormales -1);
	    region->infos3d.attributes[VAR_NZ] 
	      /= (region->infos3d.nbNormales -1);
	  }
	}
      }
    }
  }


  /* Update the various global informations` on the regionMap */
  rMap_endUpdate (regionMap);

  return (RMAP_OK);
}

/**********************************************************************
 *
 * Updates the 3D information structure with a 3D point
 *
 **********************************************************************/
void
rmap_update3dAttributes (RMAP_3D_INFOS*	infos3d,
			 DATA_PT3D*	pt3d)
{
  int		nbPts;
  double	mean;
  double	z;
  int		nbNormales;
  double	nx, ny, nz;

  z = pt3d->coord_3;
  nbPts = infos3d->attributes[NB_PTS];

  /* Update the elevation attributes */
  if (nbPts == 0) {
    infos3d->attributes[MIN_Z] = z;
    infos3d->attributes[MAX_Z] = z;
    infos3d->attributes[DELTA_Z] = 0.0;
    infos3d->attributes[MEAN_Z] = z;
    infos3d->attributes[VAR_Z] = 0.0;
  }
  else {
    infos3d->attributes[MIN_Z] = MIN (infos3d->attributes[MIN_Z], z);
    infos3d->attributes[MAX_Z] = MAX (infos3d->attributes[MAX_Z], z);
    infos3d->attributes[DELTA_Z] = 
      infos3d->attributes[MAX_Z] - infos3d->attributes[MIN_Z];

    /* Incremental mean and variance updates (according to Knuth's bible,
       Vol. 2, section 4.2.2). The actual variance will later be divided by the
       number of samples plus 1.  */
    mean = infos3d->attributes[MEAN_Z];
    infos3d->attributes[MEAN_Z] = ((mean * nbPts) + z) / (nbPts + 1);
    infos3d->attributes[VAR_Z] += 
      (z - mean) * (z - infos3d->attributes[MEAN_Z]);
  }

  /* Update the normal attributes */
  if (pt3d->normales == NORMALES_CALCULEES) {
    nx = pt3d->normales_tab[0];
    ny = pt3d->normales_tab[1];
    nz = pt3d->normales_tab[2];
    nbNormales = infos3d->nbNormales;
    if (nbNormales == 0) {
      infos3d->attributes[NX] = nx;
      infos3d->attributes[NY] = ny;
      infos3d->attributes[NZ] = nz;
      infos3d->attributes[VAR_NX] = 0.0;
      infos3d->attributes[VAR_NY] = 0.0;
      infos3d->attributes[VAR_NZ] = 0.0;
    }
    else {
      mean = infos3d->attributes[NX];
      infos3d->attributes[NX] = ((mean * nbNormales) + nx) / (nbNormales + 1);
      infos3d->attributes[VAR_NX] += 
	(nx - mean) * (nx - infos3d->attributes[NX]);
      mean = infos3d->attributes[NY];
      infos3d->attributes[NY] = ((mean * nbNormales) + ny) / (nbNormales + 1);
      infos3d->attributes[VAR_NY] += 
	(ny - mean) * (ny - infos3d->attributes[NY]);
      mean = infos3d->attributes[NZ];
      infos3d->attributes[NZ] = ((mean * nbNormales) + nz) / (nbNormales + 1);
      infos3d->attributes[VAR_NZ] += 
	(nz - mean) * (nz - infos3d->attributes[NZ]);
    }
    infos3d->nbNormales += 1;
  }

  infos3d->attributesState = COMPUTED;
  infos3d->attributes[NB_PTS] += 1;

  return;
}

