#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* structures, protos and errors of the lib */
#include "libregionMap.h"

static int lastUpdateDate = 0;

/**********************************************************************
 *
 * Filling of a GDHE_MNT structure from a REGION_MAP
 *
 **********************************************************************/

int
rMap_mapToGdheMnt (REGION_MAP*	rm,
		   int		mergedProbas,
		   int		growedProbas,
		   GDHE_MNT*	gdhe_dtm)


{
  int		i, j;		/* To loop on the cells */
  RMAP_REGION*	region;		/* A region of the original 'REGION_MAP' */
  GDHE_MNT_POINT*	gdhe_cell; /* A cell of the 'GDHE_MNT' */
  double	zScale = 0.01;
  double	zOrigin = 0.0;

  /* Initialisation of the gdhe_dtm */
  if (dtm_initGdheMnt (gdhe_dtm, rm->scale, rm->scale, zScale,
		       rm->nblig, rm->nbcol,
		       rm->origin2corner.euler.euler[3],
		       rm->origin2corner.euler.euler[4],
		       zOrigin, GDHE_MNT_X_COLUMN) != DTM_OK)
    return (-1);

  /* Loop on the regions */
  gdhe_cell = gdhe_dtm->points;
  for (i = 0; i < rm->nblig; i++) {
    for (j = 0; j < rm->nbcol; j++) {
      region = &(rm->regions[i * rm->nbcol + j]);

      if (region->updateState == RMAP_REGION_NEVER_PERCEIVED) {
	gdhe_cell->type = 0;
	gdhe_cell->z = 0;
      }
      else {			/* The region has been perceived */
	gdhe_cell->z = (short) (1.0 / zScale);
	if (mergedProbas != 0) {
	  gdhe_cell->type = 
	    255.0 * (region->infos3d.mergedProbas[RMAP_OBSTACLE]);
	  if (growedProbas != 0)
	    gdhe_cell->type = 
	      255.0 * (region->infos3d.growedProbas[RMAP_OBSTACLE]);
	}
	else
	  gdhe_cell->type = 
	    255.0 * (region->infos3d.newProbas[RMAP_OBSTACLE]);
      }
      gdhe_cell++;
    }
  }
  
  return (RMAP_OK);
}

/**********************************************************************
 *
 * Filling of a GDHE_MNT structure from a regionMap, but considering only the
 * last modified cells of the regionMap.
 *
 **********************************************************************/

int
rMap_lastMapToGdheMnt (REGION_MAP*	map,
		       int		mergedProbas,
		       int		growedProbas,
		       GDHE_MNT*	gdhe_dtm)


{
  int		i, j;		/* To loop on the cells */
  double	xOrigin;
  double	yOrigin;	/* Coordinates of the top-left corner of the
				   top-left sub-map in the global frame */
  int		lastDate;	/* date of the last map update */
  int		iMin, iMax;
  int		jMin, jMax;	/* Bounds of the last map update */
  int		nblig_gdhe;
  int		nbcol_gdhe;	/* Size of the gdhe_dtm */
  RMAP_REGION*	region;		/* A region of the original map */
  GDHE_MNT_POINT*	gdhe_cell; /* A cell of the 'GDHE_MNT' */
  GDHE_MNT_STORAGE_TYPE	storage;
  double	zScale = 0.01;
  double	zOrigin = 0.0;
  double	zOffset = 0.0;

  /* Compute the dtm bounds of the last perception, in order to determine the
     size of the gdhe_dtm. Also, find the dtm min elevation in the last
     perception */ 
  lastDate = map->nbUpdates;
  iMin = map->nblig;
  iMax = 0;
  jMin = map->nbcol;
  jMax = 0;
  region = map->regions;
  for (i = 0; i < map->nblig; i++) {
    for (j = 0; j < map->nbcol; j++) {
      if ((region->updateState != RMAP_REGION_NEVER_PERCEIVED) 
	  && (region->lastDate >= lastUpdateDate)) {
	iMin = MIN (i, iMin);
	iMax = MAX (i, iMax);
	jMin = MIN (j, jMin);
	jMax = MAX (j, jMax);
      }
      region++;
    }
  }
  nblig_gdhe = iMax - iMin + 1;
  nbcol_gdhe = jMax - jMin + 1;
  if ((nblig_gdhe <= 0) || (nbcol_gdhe <= 0)) {
    /* Vicious case where no pixels has been fused in the dtm (it can happen,
       who knows ?) */
    return rMapError (ERR_RMAP_NO_CELLS_UPDATED);
  }

  storage = GDHE_MNT_X_ROW;
  
  /* Initialisation of the gdhe_dtm. */
  xOrigin = iMin * map->scale + map->origin2corner.euler.euler[3];
  yOrigin = jMin * map->scale + map->origin2corner.euler.euler[4];
  zOrigin = - zOffset;  

  if (dtm_initGdheMnt (gdhe_dtm, map->scale, map->scale, zScale,
		       nblig_gdhe, nbcol_gdhe,
		       xOrigin, yOrigin, zOrigin,
		       storage) != DTM_OK)
    return (-1);
  
  /* Loop on the cells of the last perceptions */
  for (i = 0; i < nblig_gdhe; i++) {
    for (j = 0; j < nbcol_gdhe; j++) {
      region = map->regions + ((i + iMin) * map->nbcol) + (j + jMin);
      gdhe_cell = gdhe_dtm->points + (i * (gdhe_dtm->hdr).ncols) + j;
      if ((region->updateState != RMAP_REGION_NEVER_PERCEIVED) 
	  && (region->lastDate >= lastUpdateDate)) {
	gdhe_cell->z = (short) (1.0 / zScale);
	if (mergedProbas != 0) {
	  gdhe_cell->type = 
	    255.0 * (region->infos3d.mergedProbas[RMAP_OBSTACLE]);
	  if (growedProbas != 0)
	    gdhe_cell->type = 
	      255.0 * (region->infos3d.growedProbas[RMAP_OBSTACLE]);
	}
	else
	  gdhe_cell->type = 
	    255.0 * (region->infos3d.newProbas[RMAP_OBSTACLE]);
      }
    }
  }
  lastUpdateDate = lastDate;
  
  return (RMAP_OK);
}
