/* Generic C includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Structures, protos and erreurs of the lib */
#include "libregionMap.h"
 
/**********************************************************************
 *
 * Writes a REGION_MAP structure in a file
 *
 **********************************************************************/

int
rMap_writeAsciiMap (REGION_MAP*	regionMap,
		    FILE*	file)

{
  RMAP_REGION*		region;
  int			i, j;

  /* The keystring */
  fprintf (file, RMAP_ASCII_FF);
  fprintf (file, "\n");

  /* Writes the header informations */
  fprintf (file, "%e %d %d\n",
	   regionMap->scale, 
	   regionMap->nblig, regionMap->nbcol);
  fprintf (file, "%e %e %e %e %e\n",
	   regionMap->origin2corner.euler.euler[3], 
	   regionMap->origin2corner.euler.euler[4], 
	   regionMap->origin2corner.euler.euler[5], 
	   regionMap->size_x, regionMap->size_y);
  fprintf (file, "%e %d\n", regionMap->robotRadius, regionMap->robotCellRadius);

  fprintf (file, "%d %d %d\n",
	   regionMap->lastCellStates,  regionMap->bounds,
	   regionMap->geoOrientation);
  fprintf (file, "%d %d\n",
	   regionMap->iOrigin, regionMap->jOrigin);
  fprintf (file, "%d %d %d %d\n",
	    regionMap->last_iMin, regionMap->last_iMax, 
	    regionMap->last_jMin, regionMap->last_jMax);
  fprintf (file, "%d %d %d %d\n",
	    regionMap->iMin, regionMap->iMax, 
	    regionMap->jMin, regionMap->jMax);
  fprintf (file, "%d %d %d\n", regionMap->nbUpdates, 
	   regionMap->nbNewCells, regionMap->nbUpdatedCells);
  fprintf (file, "%d\n", regionMap->nbPointsMin);
  for (i = 0; i < NB_3D_CLASSES; i++) {
    fprintf (file, "%d ", regionMap->classes3d[i]);
  }
  fprintf (file, "\n");
  for (i = 0; i < NB_TEXTURE_CLASSES; i++) {
    fprintf (file, "%d ", regionMap->classesTexture[i]);
  }
  fprintf (file, "\n");

  /* Writes the regions */
  region = regionMap->regions;
  for (i = 0; i < regionMap->nblig; i++) {
    for (j = 0; j < regionMap->nbcol; j++) {
      rMap_writeAsciiRegion (region, file);
      region++;
    }
    fprintf (file, "\n");
  }
  
  return (RMAP_OK);
}

/**********************************************************************
 *
 * Reads a REGION_MAP structure from a file
 *
 **********************************************************************/

int
rMap_readAsciiMap (FILE*	file,
		   REGION_MAP*	regionMap)
{
  RMAP_REGION*		region;
  char			keyString[256];
  int			i, j;

  /* The key string */
  fscanf (file, "%s", keyString);
  if (strcmp (keyString, RMAP_ASCII_FF) != 0) 
    return rMapError  (ERR_RMAP_FORMAT_ASCII_FILE);

  /* Read the header informations */
  if (fscanf (file, "%le %d %d\n",
	      &(regionMap->scale), 
	      &(regionMap->nblig), &(regionMap->nbcol)) != 3)
    return rMapError  (ERR_RMAP_READ_FILE);
  if (fscanf (file, "%le %le %le %le %le\n",
	      &(regionMap->origin2corner.euler.euler[3]), 
	      &(regionMap->origin2corner.euler.euler[4]), 
	      &(regionMap->origin2corner.euler.euler[5]), 
	      &(regionMap->size_x), &(regionMap->size_y)) != 5)
    return rMapError  (ERR_RMAP_READ_FILE);
  if (fscanf (file, "%le %d\n", 
	      &(regionMap->robotRadius), 
	      &(regionMap->robotCellRadius)) != 2)
    return rMapError  (ERR_RMAP_READ_FILE);

  if (rMap_initMap (regionMap, regionMap->scale,
		    regionMap->nblig, regionMap->nbcol,
		    regionMap->origin2corner.euler.euler[3], 
		    regionMap->origin2corner.euler.euler[4], 
		    regionMap->origin2corner.euler.euler[5], 
		    regionMap->robotRadius) != RMAP_OK)
    return rMapError  (ERR_RMAP_READ_FILE);

  if (fscanf (file, "%d %d %d\n",
	      (int*) &(regionMap->lastCellStates), 
	      (int*) &(regionMap->bounds),
	      (int*) &(regionMap->geoOrientation)) != 3)
    return rMapError  (ERR_RMAP_READ_FILE);
  if (fscanf (file, "%d %d\n",
	      &(regionMap->iOrigin), &(regionMap->jOrigin)) != 2)
    return rMapError  (ERR_RMAP_READ_FILE);
  if (fscanf (file, "%d %d %d %d\n",
	      &(regionMap->last_iMin), &(regionMap->last_iMax), 
	      &(regionMap->last_jMin), &(regionMap->last_jMax)) != 4)
    return rMapError  (ERR_RMAP_READ_FILE);
  if (fscanf (file, "%d %d %d %d\n",
	      &(regionMap->iMin), &(regionMap->iMax), 
	      &(regionMap->jMin), &(regionMap->jMax)) != 4)
    return rMapError  (ERR_RMAP_READ_FILE);
  if (fscanf (file, "%d %d %d\n", &(regionMap->nbUpdates),
	      &(regionMap->nbNewCells), &(regionMap->nbUpdatedCells)) != 3)
    return rMapError  (ERR_RMAP_READ_FILE);
  if (fscanf (file, "%d\n", &(regionMap->nbPointsMin)) != 1)
    return rMapError  (ERR_RMAP_READ_FILE);
  for (i = 0; i < NB_3D_CLASSES; i++) {
    if (fscanf (file, "%d ", (int*) &(regionMap->classes3d[i])) != 1)
      return rMapError  (ERR_RMAP_READ_FILE);
  }
  fscanf (file, "\n");
  for (i = 0; i < NB_TEXTURE_CLASSES; i++) {
    if (fscanf (file, "%d ", (int*) &(regionMap->classesTexture[i])) != 1)
      return rMapError  (ERR_RMAP_READ_FILE);
  }
  fscanf (file, "\n");

  /* Read the regions */
  region = regionMap->regions;
  for (i = 0; i < regionMap->nblig; i++) {
    for (j = 0; j < regionMap->nbcol; j++) {
      if (rMap_readAsciiRegion (region, file) != RMAP_OK)
	return rMapError  (ERR_RMAP_READ_FILE);
      region++;
    }
    fscanf (file, "\n");
  }

  return (RMAP_OK);
}

/**********************************************************************
 *
 * Writes a RMAP_REGION structure in a file
 *
 **********************************************************************/
int
rMap_writeAsciiRegion (RMAP_REGION*	region,
		       FILE*		file)

{
  int	i;
  /* writes all the fields of the region */
  fprintf (file, "%d %d %d %d\n",
	   region->updateState, region->firstDate, region->lastDate, 
	   region->nbUpdates); 
  for (i = 0; i < NB_CLASSIFICATION_TYPE; i++) 
    fprintf (file, "%d ", region->classificationType[i]);

  /* 3D related informations */
  fprintf (file, "%d %d %d %d\n",
	   (region->infos3d.attributesState), 
	   (region->infos3d.probaState), 
	   (region->infos3d.classificationState), 
	   (region->infos3d.nbNormales));
  for (i = 0; i < NB_3D_ATTRIBUTES; i++)
    fprintf (file, "%e ", region->infos3d.attributes[i]);
  fprintf (file, "\n%d\n", region->infos3d.classificationState);
  for (i = 0; i < NB_3D_CLASSES; i++) {
    fprintf (file, "%e %e %e ", 
	     region->infos3d.newProbas[i], 
	     region->infos3d.mergedProbas[i],
	     region->infos3d.growedProbas[i]);
  }
  fprintf (file, "%d\n", region->infos3d.label);

  /* TODO: texture related informations */

  return (RMAP_OK);
}

/**********************************************************************
 *
 * Reads a RMAP_REGION structure in a file
 *
 **********************************************************************/
int
rMap_readAsciiRegion (RMAP_REGION*	region,
		      FILE*		file)

{
  int	i;
  /* read all the fields of the region */
  if (fscanf (file, "%d %d %d %d\n",
	      (int*) &(region->updateState), &(region->firstDate),
	      &(region->lastDate), &(region->nbUpdates)) != 4) 
    return rMapError  (ERR_RMAP_READ_FILE);
  for (i = 0; i < NB_CLASSIFICATION_TYPE; i++) 
    if (fscanf (file, "%d ", (int*) &(region->classificationType[i])) != 1)
      return rMapError  (ERR_RMAP_READ_FILE);
      
  /* 3D related informations */
  if (fscanf (file, "%d %d %d %d\n",
	      (int*) &(region->infos3d.attributesState), 
 	      (int*) &(region->infos3d.probaState),  
 	      (int*) &(region->infos3d.classificationState),  
	      &(region->infos3d.nbNormales)) != 4)
    return rMapError  (ERR_RMAP_READ_FILE);
  for (i = 0; i < NB_3D_ATTRIBUTES; i++) {
    if (fscanf (file, "%le ", &(region->infos3d.attributes[i])) != 1)
      return rMapError  (ERR_RMAP_READ_FILE);
  } 
  if (fscanf (file, "\n%d\n", 
	      (int*) &(region->infos3d.classificationState)) != 1)
    return rMapError  (ERR_RMAP_READ_FILE);
  for (i = 0; i < NB_3D_CLASSES; i++)
    if (fscanf (file, "%le %le %le ",
		&(region->infos3d.newProbas[i]),
		&(region->infos3d.mergedProbas[i]),
		&(region->infos3d.growedProbas[i])) != 3)
      return rMapError  (ERR_RMAP_READ_FILE);
  if (fscanf (file, "%d\n", (int*) &(region->infos3d.label)) != 1)
    return rMapError  (ERR_RMAP_READ_FILE);
  
  return (RMAP_OK);
}

/**********************************************************************
 *
 * Writes a RMAP_3D_PROTO_SPACE structure in a file
 *
 **********************************************************************/

int
rMap_write3DProtoSpace (RMAP_3D_PROTO_SPACE*	ps,
			FILE*			file)

{
  int i;

  /* The keystring */
  fprintf (file, RMAP_PROTO_SPACE_FF);
  fprintf (file, "\n");
  
  /* Writes the header informations */
  fprintf (file, "%d %d\n", ps->state, ps->norm);
  fprintf (file, "%d\n", ps->nbRegionMap);
  for (i = 0; i < ps->nbRegionMap; i++) 
    fprintf (file, "%s\n", ps->regionMapFiles[i]);
  fprintf (file, "%d %d\n", ps->nbProtos, ps->nbNeighbors);
  for (i = 0; i < NB_3D_CLASSES; i++) 
    fprintf (file, "%d ", ps->classes[i]);
  fprintf (file, "\n");
  for (i = 0; i < NB_3D_CLASSES; i++) 
    fprintf (file, "%d ", ps->nbPerClass[i]);
  fprintf (file, "\n");
  for (i = 0; i < NB_3D_ATTRIBUTES; i++) 
    fprintf (file, "%d ", ps->attributeMask[i]);
  fprintf (file, "\n");
  for (i = 0; i < NB_3D_ATTRIBUTES; i++) 
    fprintf (file, "%e ", ps->minValues[i]);
  fprintf (file, "\n");
  for (i = 0; i < NB_3D_ATTRIBUTES; i++) 
    fprintf (file, "%e ", ps->maxValues[i]);
  fprintf (file, "\n");
  for (i = 0; i < NB_3D_ATTRIBUTES; i++) 
    fprintf (file, "%e ", ps->deltaValues[i]);
  fprintf (file, "\n");
  for (i = 0; i < NB_3D_ATTRIBUTES; i++) 
    fprintf (file, "%e ", ps->mean[i]);
  fprintf (file, "\n");
  for (i = 0; i < NB_3D_ATTRIBUTES; i++) 
    fprintf (file, "%e ", ps->var[i]);
  fprintf (file, "\n");
  /* Write all the prototypes */
  for (i = 0; i < ps->nbProtos; i++) 
    rMap_writeAsciiRegion (&(ps->prototypes[i]), file);
  fprintf (file, "\n");
  
  return (RMAP_OK);
}

/**********************************************************************
 *
 * Reads a RMAP_3D_PROTO_SPACE structure in a file
 *
 **********************************************************************/

int
rMap_read3DProtoSpace (FILE*			file,
		       RMAP_3D_PROTO_SPACE*	ps)
  
{
  RMAP_REGION		region;
  char			keyString[256];
  int			i;

  /* The keystring */
  fscanf (file, "%s", keyString);
  if (strcmp (keyString, RMAP_PROTO_SPACE_FF) != 0) 
    return rMapError  (ERR_RMAP_FORMAT_ASCII_FILE);

  /* Read the header informations */
  if (fscanf (file, "%d %d\n", &(ps->state), &(ps->norm)) != 2)
    return rMapError  (ERR_RMAP_READ_FILE);
  if (fscanf (file, "%d\n", &(ps->nbRegionMap)) != 1)
    return rMapError  (ERR_RMAP_READ_FILE);
  for (i = 0; i < ps->nbRegionMap; i++) 
    if (fscanf (file, "%s", &(ps->regionMapFiles[i])) != 1)
      return rMapError  (ERR_RMAP_READ_FILE);
  if (fscanf (file, "%d %d\n", &(ps->nbProtos), &(ps->nbNeighbors)) != 2)
    return rMapError  (ERR_RMAP_READ_FILE);
  for (i = 0; i < NB_3D_CLASSES; i++) 
    if (fscanf (file, "%d ", &(ps->classes[i])) != 1)
      return rMapError  (ERR_RMAP_READ_FILE);
  fscanf (file, "\n");
  for (i = 0; i < NB_3D_CLASSES; i++) 
    if (fscanf (file, "%d ", &(ps->nbPerClass[i])) != 1)
      return rMapError  (ERR_RMAP_READ_FILE);
  fscanf (file, "\n");
  for (i = 0; i < NB_3D_ATTRIBUTES; i++) 
    if (fscanf (file, "%d ", &(ps->attributeMask[i])) != 1)
      return rMapError  (ERR_RMAP_READ_FILE);
  fscanf (file, "\n");
  for (i = 0; i < NB_3D_ATTRIBUTES; i++) 
    if (fscanf (file, "%le ", &(ps->minValues[i])) != 1)
      return rMapError  (ERR_RMAP_READ_FILE);
  fscanf (file, "\n");
  for (i = 0; i < NB_3D_ATTRIBUTES; i++) 
    if (fscanf (file, "%le ", &(ps->maxValues[i])) != 1)
      return rMapError  (ERR_RMAP_READ_FILE);
  fscanf (file, "\n");
  for (i = 0; i < NB_3D_ATTRIBUTES; i++) 
    if (fscanf (file, "%le ", &(ps->deltaValues[i])) != 1)
      return rMapError  (ERR_RMAP_READ_FILE);
  fscanf (file, "\n");
  for (i = 0; i < NB_3D_ATTRIBUTES; i++) 
    if (fscanf (file, "%le ", &(ps->mean[i])) != 1)
      return rMapError  (ERR_RMAP_READ_FILE);
  fscanf (file, "\n");
  for (i = 0; i < NB_3D_ATTRIBUTES; i++) 
    if (fscanf (file, "%le ", &(ps->var[i])) != 1)
      return rMapError  (ERR_RMAP_READ_FILE);
  fscanf (file, "\n");
  /* Read all the prototypes */
  for (i = 0; i < ps->nbProtos; i++) 
    if (rMap_readAsciiRegion (&region, file) != RMAP_OK)
      return rMapError  (ERR_RMAP_READ_FILE);
    else
      memcpy (&(ps->prototypes[i]), &region, sizeof (RMAP_REGION));
  fscanf (file, "\n");
  
  return (RMAP_OK);
}
